@@ -1,9 +1,8 @@
-import pluthon as plt
-import uplc
-
 from .rewrite.rewrite_augassign import RewriteAugAssign
+from .rewrite.rewrite_forbidden_overwrites import RewriteForbiddenOverwrites
 from .rewrite.rewrite_import import RewriteImport
 from .rewrite.rewrite_import_dataclasses import RewriteImportDataclasses
+from .rewrite.rewrite_import_hashlib import RewriteImportHashlib
 from .rewrite.rewrite_import_plutusdata import RewriteImportPlutusData
 from .rewrite.rewrite_import_typing import RewriteImportTyping
 from .rewrite.rewrite_inject_builtins import RewriteInjectBuiltins
@@ -14,6 +13,8 @@ from .optimize.optimize_remove_deadvars import OptimizeRemoveDeadvars
 from .optimize.optimize_varlen import OptimizeVarlen
 from .type_inference import *
 from .util import RawPlutoExpr, TypedNodeTransformer
+from .typed_ast import transform_ext_params_map, transform_output_map
+
 
 STATEMONAD = "s"
 
@@ -74,98 +75,6 @@ CmpMap = {
     },
 }
 
-EmptyListMap = {
-    IntegerInstanceType: plt.EmptyIntegerList(),
-    ByteStringInstanceType: plt.EmptyByteStringList(),
-    StringInstanceType: plt.EmptyTextList(),
-    UnitInstanceType: plt.EmptyUnitList(),
-    BoolInstanceType: plt.EmptyBoolList(),
-}
-
-
-def empty_list(p: Type):
-    if p in EmptyListMap:
-        return EmptyListMap[p]
-    assert isinstance(p, InstanceType), "Can only create lists of instances"
-    if isinstance(p.typ, ListType):
-        el = empty_list(p.typ.typ)
-        return plt.EmptyListList(uplc.BuiltinList([], el.sample_value))
-    if isinstance(p.typ, DictType):
-        el_key = empty_list(p.typ.key_typ)
-        el_value = empty_list(p.typ.value_typ)
-        return plt.EmptyListList(
-            uplc.BuiltinList(
-                [], uplc.BuiltinPair(el_key.sample_value, el_value.sample_value)
-            )
-        )
-    if isinstance(p.typ, RecordType):
-        return plt.EmptyDataList()
-    raise NotImplementedError(f"Empty lists of type {p} can't be constructed yet")
-
-
-TransformExtParamsMap = {
-    IntegerInstanceType: lambda x: plt.UnIData(x),
-    ByteStringInstanceType: lambda x: plt.UnBData(x),
-    StringInstanceType: lambda x: plt.DecodeUtf8(plt.UnBData(x)),
-    UnitInstanceType: lambda x: plt.Lambda(["_"], plt.Unit()),
-    BoolInstanceType: lambda x: plt.NotEqualsInteger(x, plt.Integer(0)),
-}
-
-
-def transform_ext_params_map(p: Type):
-    assert isinstance(
-        p, InstanceType
-    ), "Can only transform instances, not classes as input"
-    if p in TransformExtParamsMap:
-        return TransformExtParamsMap[p]
-    if isinstance(p.typ, ListType):
-        list_int_typ = p.typ.typ
-        return lambda x: plt.MapList(
-            plt.UnListData(x),
-            plt.Lambda(["x"], transform_ext_params_map(list_int_typ)(plt.Var("x"))),
-            empty_list(p.typ.typ),
-        )
-    if isinstance(p.typ, DictType):
-        # TODO also remap in the style the list is mapped (but on pairs)
-        raise NotImplementedError(
-            "Dictionaries can currently not be parsed from PlutusData"
-        )
-    return lambda x: x
-
-
-TransformOutputMap = {
-    StringInstanceType: lambda x: plt.BData(plt.EncodeUtf8(x)),
-    IntegerInstanceType: lambda x: plt.IData(x),
-    ByteStringInstanceType: lambda x: plt.BData(x),
-    UnitInstanceType: lambda x: plt.Lambda(["_"], plt.Unit()),
-    BoolInstanceType: lambda x: plt.IData(
-        plt.IfThenElse(x, plt.Integer(1), plt.Integer(0))
-    ),
-}
-
-
-def transform_output_map(p: Type):
-    assert isinstance(
-        p, InstanceType
-    ), "Can only transform instances, not classes as input"
-    if p in TransformOutputMap:
-        return TransformOutputMap[p]
-    if isinstance(p.typ, ListType):
-        list_int_typ = p.typ.typ
-        return lambda x: plt.ListData(
-            plt.MapList(
-                x,
-                plt.Lambda(["x"], transform_output_map(list_int_typ)(plt.Var("x"))),
-            ),
-        )
-    if isinstance(p.typ, DictType):
-        # TODO also remap in the style the list is mapped as input
-        raise NotImplementedError(
-            "Dictionaries can currently not be mapped to PlutusData"
-        )
-    return lambda x: x
-
-
 ConstantMap = {
     str: plt.Text,
     bytes: lambda x: plt.ByteString(x),
@@ -326,16 +235,22 @@ class UPLCCompiler(TypedNodeTransformer):
 
     def visit_Name(self, node: TypedName) -> plt.AST:
         # depending on load or store context, return the value of the variable or its name
-        if isinstance(node.ctx, Load):
+        if not isinstance(node.ctx, Load):
+            raise NotImplementedError(f"Context {node.ctx} not supported")
+        if isinstance(node.typ, ClassType):
+            # if this is not an instance but a class, call the constructor
             return plt.Lambda(
                 [STATEMONAD],
-                plt.FunctionalMapAccess(
-                    plt.Var(STATEMONAD),
-                    plt.ByteString(node.id),
-                    plt.TraceError(f"NameError: {node.orig_id}"),
-                ),
+                node.typ.constr(),
             )
-        raise NotImplementedError(f"Context {node.ctx} not supported")
+        return plt.Lambda(
+            [STATEMONAD],
+            plt.FunctionalMapAccess(
+                plt.Var(STATEMONAD),
+                plt.ByteString(node.id),
+                plt.TraceError(f"NameError: {node.orig_id}"),
+            ),
+        )
 
     def visit_Expr(self, node: TypedExpr) -> plt.AST:
         # we exploit UPLCs eager evaluation here
@@ -352,10 +267,18 @@ class UPLCCompiler(TypedNodeTransformer):
     def visit_Call(self, node: TypedCall) -> plt.AST:
         # compiled_args = " ".join(f"({self.visit(a)} {STATEMONAD})" for a in node.args)
         # return rf"(\{STATEMONAD} -> ({self.visit(node.func)} {compiled_args})"
+        # TODO function is actually not of type polymorphic function type here anymore
+        if isinstance(node.func.typ, PolymorphicFunctionInstanceType):
+            # edge case for weird builtins that are polymorphic
+            func_plt = node.func.typ.polymorphic_function.impl_from_args(
+                node.func.typ.typ.argtyps
+            )
+        else:
+            func_plt = plt.Apply(self.visit(node.func), plt.Var(STATEMONAD))
         return plt.Lambda(
             [STATEMONAD],
             plt.Apply(
-                plt.Apply(self.visit(node.func), plt.Var(STATEMONAD)),
+                func_plt,
                 # pass in all arguments evaluated with the statemonad
                 *(plt.Apply(self.visit(a), plt.Var(STATEMONAD)) for a in node.args),
                 # eventually pass in the state monad as well
@@ -486,13 +409,13 @@ class UPLCCompiler(TypedNodeTransformer):
         return self.visit_sequence([])
 
     def visit_Subscript(self, node: TypedSubscript) -> plt.AST:
-        assert isinstance(
-            node.slice, Index
-        ), "Only single index slices are currently supported"
         assert isinstance(
             node.value.typ, InstanceType
         ), "Can only access elements of instances, not classes"
         if isinstance(node.value.typ.typ, TupleType):
+            assert isinstance(
+                node.slice, Index
+            ), "Only single index slices for tuples are currently supported"
             assert isinstance(
                 node.slice.value, Constant
             ), "Only constant index access for tuples is supported"
@@ -506,6 +429,9 @@ class UPLCCompiler(TypedNodeTransformer):
                 ),
             )
         if isinstance(node.value.typ.typ, ListType):
+            assert isinstance(
+                node.slice, Index
+            ), "Only single index slices for lists are currently supported"
             assert (
                 node.slice.value.typ == IntegerInstanceType
             ), "Only single element list index access supported"
@@ -516,6 +442,29 @@ class UPLCCompiler(TypedNodeTransformer):
                     plt.Apply(self.visit(node.slice.value), plt.Var(STATEMONAD)),
                 ),
             )
+        elif isinstance(node.value.typ.typ, ByteStringType):
+            if isinstance(node.slice, Index):
+                return plt.Lambda(
+                    [STATEMONAD],
+                    plt.IndexByteString(
+                        plt.Apply(self.visit(node.value), plt.Var(STATEMONAD)),
+                        plt.Apply(self.visit(node.slice.value), plt.Var(STATEMONAD)),
+                    ),
+                )
+            elif isinstance(node.slice, Slice):
+                return plt.Lambda(
+                    [STATEMONAD],
+                    plt.SliceByteString(
+                        plt.Apply(self.visit(node.slice.lower), plt.Var(STATEMONAD)),
+                        plt.SubtractInteger(
+                            plt.Apply(
+                                self.visit(node.slice.upper), plt.Var(STATEMONAD)
+                            ),
+                            plt.Integer(1),
+                        ),
+                        plt.Apply(self.visit(node.value), plt.Var(STATEMONAD)),
+                    ),
+                )
         raise NotImplementedError(f"Could not implement subscript of {node}")
 
     def visit_Tuple(self, node: TypedTuple) -> plt.AST:
@@ -531,21 +480,13 @@ class UPLCCompiler(TypedNodeTransformer):
         return self.visit_sequence([])
 
     def visit_Attribute(self, node: TypedAttribute) -> plt.AST:
-        if node.pos == -1:
-            # access to constructor
-            return plt.Lambda(
-                [STATEMONAD],
-                plt.Constructor(plt.Apply(self.visit(node.value), plt.Var(STATEMONAD))),
-            )
-        # access to normal fields
+        assert isinstance(
+            node.typ, InstanceType
+        ), "Can only access attributes of instances"
+        obj = self.visit(node.value)
+        attr = node.value.typ.attribute(node.attr)
         return plt.Lambda(
-            [STATEMONAD],
-            transform_ext_params_map(node.typ)(
-                plt.NthField(
-                    plt.Apply(self.visit(node.value), plt.Var(STATEMONAD)),
-                    plt.Integer(node.pos),
-                ),
-            ),
+            [STATEMONAD], plt.Apply(attr, plt.Apply(obj, plt.Var(STATEMONAD)))
         )
 
     def visit_Assert(self, node: TypedAssert) -> plt.AST:
@@ -568,6 +509,14 @@ class UPLCCompiler(TypedNodeTransformer):
     def visit_RawPlutoExpr(self, node: RawPlutoExpr) -> plt.AST:
         return node.expr
 
+    def visit_List(self, node: TypedList) -> plt.AST:
+        assert isinstance(node.typ, InstanceType)
+        assert isinstance(node.typ.typ, ListType)
+        l = empty_list(node.typ.typ.typ)
+        for e in node.elts:
+            l = plt.MkCons(plt.Apply(self.visit(e), plt.Var(STATEMONAD)), l)
+        return plt.Lambda([STATEMONAD], l)
+
     def generic_visit(self, node: AST) -> plt.AST:
         raise NotImplementedError(f"Can not compile {node}")
 
@@ -580,7 +529,9 @@ def compile(prog: AST):
         RewriteAugAssign,
         RewriteTupleAssign,
         RewriteImportPlutusData,
+        RewriteImportHashlib,
         RewriteImportTyping,
+        RewriteForbiddenOverwrites,
         RewriteImportDataclasses,
         # The type inference needs to be run after complex python operations were rewritten
         AggressiveTypeInferencer,
