@@ -24,15 +24,8 @@ security into the Smart Contract by checking type correctness.
 INITIAL_SCOPE = dict(
     {
         # class annotations
-        "bytes": ByteStringType,
-        "int": IntegerType,
-        # just to block overwriting
-        "List": ListType(NoneInstanceType),
-        "Dict": DictType(NoneInstanceType, NoneInstanceType),
-        "Optional": OptionalType(NoneInstanceType),
-        "Union": UnionType(NoneInstanceType),
-        "dataclass": NoneInstanceType,
-        "PlutusData": NoneInstanceType,
+        "bytes": ByteStringType(),
+        "int": IntegerType(),
     }
 )
 
@@ -79,10 +72,10 @@ class AggressiveTypeInferencer(NodeTransformer):
         if isinstance(ann, Tuple):
             if not ann.elts:
                 # This is ()
-                return UnitType
+                return UnitType()
         if isinstance(ann, Name):
             if ann.id in ATOMIC_TYPES:
-                return AtomicType(ann.id)
+                return ATOMIC_TYPES[ann.id]
             v_t = self.variable_type(ann.id)
             if isinstance(v_t, ClassType):
                 return v_t
@@ -92,22 +85,16 @@ class AggressiveTypeInferencer(NodeTransformer):
         if isinstance(ann, Subscript):
             assert isinstance(
                 ann.value, Name
-            ), "Only Optional, Union, Dict and List are allowed as Generic types"
+            ), "Only Union, Dict and List are allowed as Generic types"
             assert isinstance(ann.slice, Index), "Generic types must be parameterized"
-            if ann.value.id == "Optional":
-                ann_type = self.type_from_annotation(ann.slice.value)
-                assert isinstance(
-                    ann_type, ClassType
-                ), "Optional must have a single type as parameter"
-                return OptionalType(ann_type)
             if ann.value.id == "Union":
                 assert isinstance(
                     ann.slice.value, Tuple
                 ), "Union must combine multiple classes"
                 ann_types = [self.type_from_annotation(e) for e in ann.slice.value.elts]
                 assert all(
-                    isinstance(e, ClassType) for e in ann_types
-                ), "Union must combine multiple classes"
+                    isinstance(e, RecordType) for e in ann_types
+                ), "Union must combine multiple PlutusData classes"
                 return UnionType(FrozenFrozenList(FrozenFrozenList(ann_types)))
             if ann.value.id == "List":
                 ann_type = self.type_from_annotation(ann.slice.value)
@@ -128,7 +115,7 @@ class AggressiveTypeInferencer(NodeTransformer):
                 ), "Dict must combine two classes"
                 return DictType(*(InstanceType(a) for a in ann_types))
             raise NotImplementedError(
-                "Only Optional, Union, Dict and List are allowed as Generic types"
+                "Only Union, Dict and List are allowed as Generic types"
             )
         if ann is None:
             TypeInferenceError(
@@ -153,7 +140,7 @@ class AggressiveTypeInferencer(NodeTransformer):
         if tc.value is None:
             tc.typ = NoneInstanceType
         else:
-            tc.typ = InstanceType(AtomicType(type(node.value).__name__))
+            tc.typ = InstanceType(ATOMIC_TYPES[type(node.value).__name__])
         return tc
 
     def visit_Tuple(self, node: Tuple) -> TypedTuple:
@@ -168,11 +155,11 @@ class AggressiveTypeInferencer(NodeTransformer):
     def visit_List(self, node: List) -> TypedList:
         tt = copy(node)
         tt.elts = [self.visit(e) for e in node.elts]
-        l_typ = tt.elts[0]
+        l_typ = tt.elts[0].typ
         assert all(
             e.typ == l_typ for e in tt.elts
         ), "All elements of a list must have the same type"
-        tt.typ = ListType(l_typ)
+        tt.typ = InstanceType(ListType(l_typ))
         return tt
 
     def visit_Assign(self, node: Assign) -> TypedAssign:
@@ -195,7 +182,7 @@ class AggressiveTypeInferencer(NodeTransformer):
             and typed_if.test.func.id == "isinstance"
         ):
             tc = typed_if.test
-            # special case for Optional and Union
+            # special case for Union
             assert isinstance(
                 tc.args[0], Name
             ), "Target 0 of an isinstance cast must be a variable name"
@@ -364,22 +351,24 @@ class AggressiveTypeInferencer(NodeTransformer):
 
     def visit_Subscript(self, node: Subscript) -> TypedSubscript:
         ts = copy(node)
-        assert isinstance(
-            ts.slice, Index
-        ), "Only single index slices are currently supported"
-        # special case: Subscript of Union / Optional / Dict / List and atomic types
+        # special case: Subscript of Union / Dict / List and atomic types
         if isinstance(ts.value, Name) and ts.value.id in [
             "Union",
-            "Optional",
             "Dict",
             "List",
         ]:
+            assert isinstance(
+                ts.slice, Index
+            ), "Only single index slices for generic types are currently supported"
             ts.value = ts.typ = self.type_from_annotation(ts)
             return ts
 
         ts.value = self.visit(node.value)
         assert isinstance(ts.value.typ, InstanceType), "Can only subscript instances"
         if isinstance(ts.value.typ.typ, TupleType):
+            assert isinstance(
+                ts.slice, Index
+            ), "Only single index slices for tuples are currently supported"
             if all(ts.value.typ.typ.typs[0] == t for t in ts.value.typ.typ.typs):
                 ts.typ = ts.value.typ.typ.typs[0]
             elif isinstance(ts.slice.value, Constant) and isinstance(
@@ -389,11 +378,37 @@ class AggressiveTypeInferencer(NodeTransformer):
             else:
                 raise TypeInferenceError(f"Could not infer type of subscript {node}")
         elif isinstance(ts.value.typ.typ, ListType):
+            assert isinstance(
+                ts.slice, Index
+            ), "Only single index slices for lists are currently supported"
             ts.typ = ts.value.typ.typ.typ
             ts.slice.value = self.visit(node.slice.value)
             assert (
                 ts.slice.value.typ == IntegerInstanceType
             ), "List indices must be integers"
+        elif isinstance(ts.value.typ.typ, ByteStringType):
+            if isinstance(ts.slice, Index):
+                ts.typ = IntegerInstanceType
+                ts.slice.value = self.visit(node.slice.value)
+                assert (
+                    ts.slice.value.typ == IntegerInstanceType
+                ), "bytes indices must be integers"
+            elif isinstance(ts.slice, Slice):
+                ts.typ = ByteStringInstanceType
+                if ts.slice.lower is None:
+                    ts.slice.lower = Constant(0)
+                ts.slice.lower = self.visit(node.slice.lower)
+                assert (
+                    ts.slice.lower.typ == IntegerInstanceType
+                ), "lower slice indices for bytes must be integers"
+                if ts.slice.upper is None:
+                    ts.slice.upper = Call(
+                        func=Name(id="len", ctx=Load()), args=[ts.value], keywords=[]
+                    )
+                ts.slice.upper = self.visit(node.slice.upper)
+                assert (
+                    ts.slice.upper.typ == IntegerInstanceType
+                ), "upper slice indices for bytes must be integers"
         else:
             raise TypeInferenceError(f"Could not infer type of subscript {node}")
         return ts
@@ -401,20 +416,31 @@ class AggressiveTypeInferencer(NodeTransformer):
     def visit_Call(self, node: Call) -> TypedCall:
         assert not node.keywords, "Keyword arguments are not supported yet"
         tc = copy(node)
+        tc.args = [self.visit(a) for a in node.args]
         tc.func = self.visit(node.func)
+        # might be a cast
         if isinstance(tc.func.typ, ClassType):
-            # TODO make sure that arguments to class init are filled
-            raise NotImplementedError("Class initialization not implemented yet")
+            tc.func.typ = tc.func.typ.constr_type()
+        # type might only turn out after the initialization (note the constr could be polymorphic)
+        if isinstance(tc.func.typ, InstanceType) and isinstance(
+            tc.func.typ.typ, PolymorphicFunctionType
+        ):
+            tc.func.typ = PolymorphicFunctionInstanceType(
+                tc.func.typ.typ.polymorphic_function.type_from_args(
+                    [a.typ for a in tc.args]
+                ),
+                tc.func.typ.typ.polymorphic_function,
+            )
         if isinstance(tc.func.typ, InstanceType) and isinstance(
             tc.func.typ.typ, FunctionType
         ):
             functyp = tc.func.typ.typ
-            tc.args = [self.visit(a) for a in node.args]
             assert len(tc.args) == len(
                 functyp.argtyps
             ), f"Signature of function {node} does not match number of arguments"
+            # all arguments need to be supertypes of the given type
             assert all(
-                a.typ == ap for a, ap in zip(tc.args, functyp.argtyps)
+                ap >= a.typ for a, ap in zip(tc.args, functyp.argtyps)
             ), f"Signature of function {node} does not match arguments"
             tc.typ = functyp.rettyp
             return tc
@@ -434,28 +460,8 @@ class AggressiveTypeInferencer(NodeTransformer):
         tp = copy(node)
         tp.value = self.visit(node.value)
         owner = tp.value.typ
-        assert isinstance(owner, InstanceType) and (
-            isinstance(owner.typ, RecordType)
-            or isinstance(owner.typ, UnionType)
-            or isinstance(owner.typ, OptionalType)
-        ), "Accessing attribute of non-instance"
-        owner_typ = owner.typ
-
-        # access to constructor
-        if tp.attr == "CONSTR_ID":
-            tp.typ = IntegerInstanceType
-            tp.pos = -1
-            return tp
         # accesses to field
-        tp.typ = None
-        for i, (attr_name, attr_type) in enumerate(owner_typ.record.fields):
-            if attr_name == tp.attr:
-                tp.typ = attr_type
-                tp.pos = i
-        if tp.typ is None:
-            assert (
-                tp.attr in tp.value.typ.record.fields
-            ), "Accessing undefined attribute of class-type"
+        tp.typ = owner.attribute_type(node.attr)
         return tp
 
     def visit_Assert(self, node: Assert) -> TypedAssert:
